/* Copyright (C) 1999 Free Software Foundation, Inc.
   This file is part of the GNU UTF-8 Library.

   The GNU UTF-8 Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   The GNU UTF-8 Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with the GNU UTF-8 Library; see the file COPYING.LIB.  If not,
   write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */

#include "utf8.h"
#include <stdio.h>

int
utf8_mblen (const char *s, size_t n)
{
  if (s == NULL)
    {
      return 0;
    }
  else
    {
      const char *start = s;
      for (;;)
	{
	  unsigned char c;
	  if (n == 0)
	    return -1;
	  c = (unsigned char) *s;
	  if (c < 0x80)
	    {
	      return (c != 0);
	    }
	  else if (c < 0xC0)
	    {
	      /* skip spurious 10XXXXXX byte */
	      s++;
	      n--;
	    }
	  else
	    {
	      size_t count;
	      int check_unsafe;
	      if (c < 0xE0)
		{
		  count = 1;
		  if (c < 0xC2)
		    goto bad_input;
		  check_unsafe = 0;
		}
	      else if (c < 0xF0)
		{
		  count = 2;
		  check_unsafe = (c == 0xE0);
		}
#if WCHAR_T_BITS == 32
	      else if (c < 0xF8)
		{
		  count = 3;
		  check_unsafe = (c == 0xF0);
		}
	      else if (c < 0xFC)
		{
		  count = 4;
		  check_unsafe = (c == 0xF8);
		}
	      else if (c < 0xFE)
		{
		  count = 5;
		  check_unsafe = (c == 0xFC);
		}
#endif
	      else
		goto bad_input;
	      if (n <= count)
		return -1;
	      s++;
	      c = (unsigned char) *s++ ^ 0x80;
	      if (!(c < 0x40))
		goto bad_input;
	      if (--count > 0)
		{
		  if (check_unsafe && ((c >> (6 - count)) == 0))
		    goto bad_input;
		  do
		    {
		      c = (unsigned char) *s++ ^ 0x80;
		      if (!(c < 0x40))
			goto bad_input;
		    }
		  while (--count > 0);
		}
	      return s - start;
	    }
	}
    }
bad_input:
  errno = EILSEQ;
  return -1;
}

size_t utf8_mbrlen (const char *s, size_t n)
{
  return utf8_mbrtowc (NULL, s, n);
}

size_t utf8_mbrtowc (wchar_t * pwc, const char *s, size_t n)
{
  if (s == NULL)
    {
      return 0;
    }
  else
    {
      const char *start = s;
      for (;;)
	{
	  unsigned char c;
	  if (n == 0)
	    return (size_t) (-2);
	  c = (unsigned char) *s;
	  if (c < 0x80)
	    {
	      if (pwc != NULL)
		*pwc = (wchar_t) c;
	      return (c != 0);
	    }
	  else if (c < 0xC0)
	    {
	      /* skip spurious 10XXXXXX byte */
	      s++;
	      n--;
	    }
	  else
	    {
	      wchar_t wc;
	      size_t count;
	      if (c < 0xE0)
		{
		  wc = (wchar_t) (c & 0x1F) << 6;
		  count = 1;
		  if (c < 0xC2)
		    goto bad_input;
		}
	      else if (c < 0xF0)
		{
		  wc = (wchar_t) (c & 0x0F) << 12;
		  count = 2;
		}
#if WCHAR_T_BITS == 32
	      else if (c < 0xF8)
		{
		  wc = (wchar_t) (c & 0x07) << 18;
		  count = 3;
		}
	      else if (c < 0xFC)
		{
		  wc = (wchar_t) (c & 0x03) << 24;
		  count = 4;
		}
	      else if (c < 0xFE)
		{
		  wc = (wchar_t) (c & 0x01) << 30;
		  count = 5;
		}
#endif
	      else
		goto bad_input;
	      if (n <= count)
		return (size_t) (-2);
	      s++;
	      c = (unsigned char) *s++ ^ 0x80;
	      if (!(c < 0x40))
		goto bad_input;
	      wc |= (wchar_t) c << (6 * --count);
	      if (count > 0)
		{
		  if ((u_wchar_t) wc < ((u_wchar_t) 1 << (5 * count + 6)))
		    goto bad_input;
		  do
		    {
		      c = (unsigned char) *s++ ^ 0x80;
		      if (!(c < 0x40))
			goto bad_input;
		      wc |= (wchar_t) c << (6 * --count);
		    }
		  while (count > 0);
		}
	      if (pwc != NULL)
		*pwc = wc;
	      return s - start;
	    }
	}
    }
bad_input:
  errno = EILSEQ;
  return (size_t) (-1);
}

size_t
utf8_mbsnrtowcs (wchar_t * dest, const char **srcp, size_t nms, size_t len)
{
  const char *src = *srcp;
  if (dest != NULL)
    {
      wchar_t *destptr = dest;
      for (; nms > 0 && len > 0; destptr++, len--)
	{
	  unsigned char c = (unsigned char) *src;
	  if (c < 0x80)
	    {
	      *destptr = (wchar_t) c;
	      if (c == 0)
		{
		  src = NULL;
		  break;
		}
	      src++;
	      nms--;
	    }
	  else if (c < 0xC0)
	    {
	      /* skip spurious 10XXXXXX byte */
	      src++;
	      nms--;
	    }
	  else
	    {
	      wchar_t wc;
	      size_t count;
	      if (c < 0xE0)
		{
		  wc = (wchar_t) (c & 0x1F) << 6;
		  count = 1;
		  if (c < 0xC2)
		    goto bad_input;
		}
	      else if (c < 0xF0)
		{
		  wc = (wchar_t) (c & 0x0F) << 12;
		  count = 2;
		}
#if WCHAR_T_BITS == 32
	      else if (c < 0xF8)
		{
		  wc = (wchar_t) (c & 0x07) << 18;
		  count = 3;
		}
	      else if (c < 0xFC)
		{
		  wc = (wchar_t) (c & 0x03) << 24;
		  count = 4;
		}
	      else if (c < 0xFE)
		{
		  wc = (wchar_t) (c & 0x01) << 30;
		  count = 5;
		}
#endif
	      else
		goto bad_input;
	      if (nms <= count)
		break;
	      nms -= count + 1;
	      src++;
	      c = (unsigned char) *src++ ^ 0x80;
	      if (!(c < 0x40))
		goto bad_input;
	      wc |= (wchar_t) c << (6 * --count);
	      if (count > 0)
		{
		  if ((u_wchar_t) wc < ((u_wchar_t) 1 << (5 * count + 6)))
		    goto bad_input;
		  do
		    {
		      c = (unsigned char) *src++ ^ 0x80;
		      if (!(c < 0x40))
			goto bad_input;
		      wc |= (wchar_t) c << (6 * --count);
		    }
		  while (count > 0);
		}
	      *destptr = wc;
	    }
	}
      *srcp = src;
      return destptr - dest;
    }
  else
    {
      /* Ignore dest and len. */
      size_t totalcount = 0;
      for (; nms > 0; totalcount++)
	{
	  unsigned char c = (unsigned char) *src;
	  if (c < 0x80)
	    {
	      if (c == 0)
		{
		  src = NULL;
		  break;
		}
	      src++;
	      nms--;
	    }
	  else if (c < 0xC0)
	    {
	      /* skip spurious 10XXXXXX byte */
	      src++;
	      nms--;
	    }
	  else
	    {
	      wchar_t wc;
	      size_t count;
	      if (c < 0xE0)
		{
		  wc = (wchar_t) (c & 0x1F) << 6;
		  count = 1;
		}
	      else if (c < 0xF0)
		{
		  wc = (wchar_t) (c & 0x0F) << 12;
		  count = 2;
		}
#if WCHAR_T_BITS == 32
	      else if (c < 0xF8)
		{
		  wc = (wchar_t) (c & 0x07) << 18;
		  count = 3;
		}
	      else if (c < 0xFC)
		{
		  wc = (wchar_t) (c & 0x03) << 24;
		  count = 4;
		}
	      else if (c < 0xFE)
		{
		  wc = (wchar_t) (c & 0x01) << 30;
		  count = 5;
		}
#endif
	      else
		goto bad_input;
	      if (nms <= count)
		break;
	      nms -= count + 1;
	      src++;
	      do
		{
		  c = (unsigned char) *src++ ^ 0x80;
		  if (!(c < 0x40))
		    goto bad_input;
		}
	      while (--count > 0);
	    }
	}
      *srcp = src;
      return totalcount;
    }
bad_input:
  /* Go back to the invalid byte. */
  --src;
  /* Skip all continuation bytes before it. */
  while (((unsigned char) *--src ^ 0x80) < 0x40);
  *srcp = src;
  return (size_t) (-1);
}

size_t utf8_mbsrtowcs (wchar_t * dest, const char **srcp, size_t len)
{
  const char *src = *srcp;
  if (dest != NULL)
    {
      wchar_t *destptr = dest;
      for (; len > 0; destptr++, len--)
	{
	  unsigned char c = (unsigned char) *src;
	  if (c < 0x80)
	    {
	      *destptr = (wchar_t) c;
	      if (c == 0)
		{
		  src = NULL;
		  break;
		}
	      src++;
	    }
	  else if (c < 0xC0)
	    {
	      /* skip spurious 10XXXXXX byte */
	      src++;
	    }
	  else
	    {
	      wchar_t wc;
	      size_t count;
	      if (c < 0xE0)
		{
		  wc = (wchar_t) (c & 0x1F) << 6;
		  count = 1;
		  if (c < 0xC2)
		    goto bad_input;
		}
	      else if (c < 0xF0)
		{
		  wc = (wchar_t) (c & 0x0F) << 12;
		  count = 2;
		}
#if WCHAR_T_BITS == 32
	      else if (c < 0xF8)
		{
		  wc = (wchar_t) (c & 0x07) << 18;
		  count = 3;
		}
	      else if (c < 0xFC)
		{
		  wc = (wchar_t) (c & 0x03) << 24;
		  count = 4;
		}
	      else if (c < 0xFE)
		{
		  wc = (wchar_t) (c & 0x01) << 30;
		  count = 5;
		}
#endif
	      else
		goto bad_input;
	      src++;
	      c = (unsigned char) *src++ ^ 0x80;
	      if (!(c < 0x40))
		goto bad_input;
	      wc |= (wchar_t) c << (6 * --count);
	      if (count > 0)
		{
		  if ((u_wchar_t) wc < ((u_wchar_t) 1 << (5 * count + 6)))
		    goto bad_input;
		  do
		    {
		      c = (unsigned char) *src++ ^ 0x80;
		      if (!(c < 0x40))
			goto bad_input;
		      wc |= (wchar_t) c << (6 * --count);
		    }
		  while (count > 0);
		}
	      *destptr = wc;
	    }
	}
      *srcp = src;
      return destptr - dest;
    }
  else
    {
      /* Ignore dest and len. */
      size_t totalcount = 0;
      for (;; totalcount++)
	{
	  unsigned char c = (unsigned char) *src;
	  if (c < 0x80)
	    {
	      if (c == 0)
		{
		  src = NULL;
		  break;
		}
	      src++;
	    }
	  else if (c < 0xC0)
	    {
	      /* skip spurious 10XXXXXX byte */
	      src++;
	    }
	  else
	    {
	      wchar_t wc;
	      size_t count;
	      if (c < 0xE0)
		{
		  wc = (wchar_t) (c & 0x1F) << 6;
		  count = 1;
		}
	      else if (c < 0xF0)
		{
		  wc = (wchar_t) (c & 0x0F) << 12;
		  count = 2;
		}
#if WCHAR_T_BITS == 32
	      else if (c < 0xF8)
		{
		  wc = (wchar_t) (c & 0x07) << 18;
		  count = 3;
		}
	      else if (c < 0xFC)
		{
		  wc = (wchar_t) (c & 0x03) << 24;
		  count = 4;
		}
	      else if (c < 0xFE)
		{
		  wc = (wchar_t) (c & 0x01) << 30;
		  count = 5;
		}
#endif
	      else
		goto bad_input;
	      src++;
	      do
		{
		  c = (unsigned char) *src++ ^ 0x80;
		  if (!(c < 0x40))
		    goto bad_input;
		}
	      while (--count > 0);
	    }
	}
      *srcp = src;
      return totalcount;
    }
bad_input:
  /* Go back to the invalid byte. */
  --src;
  /* Skip all continuation bytes before it. */
  while (((unsigned char) *--src ^ 0x80) < 0x40);
  *srcp = src;
  return (size_t) (-1);
}

int
utf8_mbtowc (wchar_t * pwc, const char *s, size_t n)
{
  /* If s == NULL, mbrtowc will return 0, which is what we want, because
     the encoding is not state dependent. */
  int result = utf8_mbrtowc (pwc, s, n);
  if (result >= 0)
    return result;
  else
    return -1;
}

wint_t utf8_fgetwc (FILE * stream)
{
  wchar_t wc;
  char buf[MB_LEN_MAX];
  size_t n;
  for (n = 0;;)
    {
      int c = getc (stream);
      if (c == EOF)
	return WEOF;
      buf[n++] = (unsigned char) c;
      switch (utf8_mbrtowc (&wc, buf, n))
	{
	case (size_t) (-2):
	  break;
	case (size_t) (-1):
	  return WEOF;
	default:
	  return wc;
	}
      if (n == MB_LEN_MAX)
	{
	  errno = EILSEQ;
	  return WEOF;
	}
    }
}

wchar_t *
utf8_fgetws (wchar_t * ws, int n, FILE * stream)
{
  wchar_t *wp;
  if (n <= 0 || ferror (stream))
    return NULL;
  wp = ws;
  while (--n > 0)
    {
      wint_t wc = utf8_getwc (stream);
      if (wc == WEOF)
	break;
      *wp++ = wc;
      if (wc == (wchar_t) '\n')
	break;
    }
  if (wp == ws || ferror (stream))
    return NULL;
  *wp = (wchar_t) '\0';
  return ws;
}

wint_t utf8_fputwc (wchar_t wc, FILE * stream)
{
  char buf[MB_LEN_MAX];
  int n = utf8_wctomb (buf, wc);
  if (n < 0)
    {
      /* FIXME: It is not clear whether we should set stream's error indicator
         here. The SingleUnix spec seems to say yes, the ISO C99 spec says no.
         Either way, we don't do it because there is no portable way to do it. */
      errno = EILSEQ;
      return WEOF;
    }
  if (fwrite (buf, n, 1, stream) < 1)
    return WEOF;
  return wc;
}

int
utf8_fputws (const wchar_t * ws, FILE * stream)
{
  for (; *ws != (wchar_t) '\0'; ws++)
    if (utf8_putwc (*ws, stream) == WEOF)
      return -1;
  return 0;
}

wint_t utf8_getwchar (void)
{
  return utf8_getwc (stdin);
}

wint_t utf8_putwchar (wchar_t wc)
{
  return utf8_putwc (wc, stdout);
}

wint_t utf8_ungetwc (wint_t wc, FILE * stream)
{
  if (wc == WEOF)
    return WEOF;
  else
    {
      char buf[MB_LEN_MAX];
      int n = utf8_wctomb (buf, wc);
      if (n < 0)
	return WEOF;
      /* glibc allows to ungetc multiple bytes in a row. */
      while (--n >= 0)
	{
	  if (ungetc (buf[n], stream) == EOF)
	    {
	      /* If it fails, try to get it right on seekable streams nevertheless. */
	      if (fseek (stream, -(n + 1), SEEK_CUR) < 0)
		return WEOF;
	      else
		break;
	    }
	}
      return wc;
    }
}

int
utf8_wctomb (char *s, wchar_t swc)
{
  u_wchar_t wc = swc;		/* since we don't know whether wchar_t is unsigned */
  size_t count;
  int result;

  if (s == NULL)
    return 0;			/* we have no state dependent encodings */

  if (wc < 0x80)
    {
      *s = (unsigned char) wc;
      return 1;
    }
  if (wc < 0x800)
    {
      count = 1;
      *s = (unsigned char) ((wc >> 6) | 0xC0);
    }
  else
#if WCHAR_T_BITS == 32
  if (wc < 0x10000)
#endif
    {
      count = 2;
      *s = (unsigned char) ((wc >> 12) | 0xE0);
    }
#if WCHAR_T_BITS == 32
  else if (wc < 0x200000)
    {
      count = 3;
      *s = (unsigned char) ((wc >> 18) | 0xF0);
    }
  else if (wc < 0x4000000)
    {
      count = 4;
      *s = (unsigned char) ((wc >> 24) | 0xF8);
    }
  else if (wc < 0x80000000)
    {
      count = 5;
      *s = (unsigned char) ((wc >> 30) | 0xFC);
    }
  else
    {
      return -1;
    }
#endif
  result = count + 1;
  do
    {
      *++s = (unsigned char) (((wc >> (6 * --count)) & 0x3F) | 0x80);
    }
  while (count > 0);
  return result;
}
