/*
 * Copyright (c) 2000 Greg Haerr <greg@censoft.com>
 * T1lib Adobe type1 routines contributed by Vidar Hokstad
 * Freetype TrueType routines contributed by Martin Jolicoeur
 *
 * Device-independent font and text drawing routines
 *
 * These routines do the necessary range checking, clipping, and cursor
 * overwriting checks, and then call the lower level device dependent
 * routines to actually do the drawing.  The lower level routines are
 * only called when it is known that all the pixels to be drawn are
 * within the device area and are visible.
 */
/*#define NDEBUG*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "device.h"
#if (UNIX | DOS_DJGPP)
#define strcmpi	strcasecmp
#endif

#if HAVE_T1LIB_SUPPORT
#include <t1lib.h>
#define T1LIB_USE_AA_HIGH

typedef struct {
	PMWFONTPROCS	procs;		/* common hdr*/
	int		defencoding;
	int		fontid;		/* t1lib stuff*/
} MWT1LIBFONT, *PMWT1LIBFONT;

static int  t1lib_init(PSD psd);
static PMWT1LIBFONT t1lib_createfont(const char *name);
static void t1lib_drawtext(PMWFONT pfont, PSD psd, COORD x, COORD y,
		const void *text, int cc, int flags);
static BOOL t1lib_getfontinfo(PMWFONT pfont, PFONTINFO pfontinfo);
static void t1lib_gettextsize(PMWFONT pfont, const void *text, int cc,
		COORD *pwidth, COORD *pheight, COORD *pbase);
static void t1lib_unloadfont(PMWFONT pfont);

/* handling routines for MWT1LIBFONT*/
static MWFONTPROCS t1lib_procs = {
	TF_ASCII,			/* routines expect ascii*/
	t1lib_getfontinfo,
	t1lib_gettextsize,
	NULL,				/* gettextbits*/
	t1lib_unloadfont,
	t1lib_drawtext
};
#endif

#ifdef T1LIB_USE_AA_HIGH
typedef unsigned long OUTPIXELVAL;
#else
typedef PIXELVAL OUTPIXELVAL;
#endif

#if HAVE_FREETYPE_SUPPORT
#include <freetype/freetype.h>
#include <freetype/ftxkern.h>
#include <math.h>

typedef struct {
	PMWFONTPROCS	procs;		/* common hdr*/
	int		defencoding;
	TT_Engine 	engine;		/* freetype stuff*/
	TT_Face 	face;
	TT_Instance	instance;
	TT_CharMap 	char_map;
	TT_Kerning 	directory;
	TT_Matrix 	matrix;
	TT_Glyph 	glyph;
	BOOL 		can_kern;
	short 		last_glyph_code;
	short 		last_pen_pos;
} MWFREETYPEFONT, *PMWFREETYPEFONT;

static PMWFREETYPEFONT freetype_createfont(const char *name);
static BOOL freetype_getfontinfo(PMWFONT pfont, PFONTINFO pfontinfo);
static void freetype_gettextsize(PMWFONT pfont, const void *text,
		int cc, COORD *pwidth, COORD *pheight, COORD *pbase);
static void freetype_unloadfont(PMWFONT pfont);
static void freetype_drawtext(PMWFONT pfont, PSD psd, COORD x, COORD y,
		const void *text, int cc, int flags);

/* handling routines for MWFREETYPEFONT*/
static MWFONTPROCS freetype_procs = {
	TF_UC16,			/* routines expect unicode 16*/
	freetype_getfontinfo,
	freetype_gettextsize,
	NULL,				/* gettextbits*/
	freetype_unloadfont,
	freetype_drawtext
};
#endif

static PMWFONT	gr_pfont;            	/* current font*/
static int	gr_fontsize = FONT_SIZE_DEFAULT;
static BOOL	gr_antialias = TRUE;
static BOOL	gr_kerning = TRUE;
static int	gr_rotation = 33;	/* rotation in degrees*/

/* temp extern decls*/
extern PIXELVAL gr_foreground;
extern PIXELVAL gr_background;
extern BOOL gr_usebg;

static void alphablend(PSD psd, OUTPIXELVAL *out, PIXELVAL src, PIXELVAL dst,
		unsigned char *alpha, int count);

/*
 * Set the font for future calls.
 */
PMWFONT
GdSetFont(PMWFONT pfont)
{
	PMWFONT	oldfont = gr_pfont;

	gr_pfont = pfont;
	return oldfont;
}

/*
 * Set the font size for future calls.
 */
COORD
GdSetFontSize(COORD fontsize)
{
	COORD	oldfontsize = gr_fontsize;

	gr_fontsize = fontsize;
	return oldfontsize;
}

/*
 * Select a font, based on various parameters
 * 
 * If height is 0, return builtin font from passed name.
 * Otherwise find builtin font best match based on height.
 */
PMWFONT
GdCreateFont(PSD psd, const char *name, COORD height/*, PMWLOGFONT plogfont*/)
{
	int 		i;
	int		fontht;
	int		fontno;
	PMWFONT		pfont;
	PMWCOREFONT	pf = psd->builtin_fonts;
	FONTINFO	fontinfo;
	SCREENINFO 	scrinfo;

	GdGetScreenInfo(psd, &scrinfo);

	/* use builtin screen fonts, FONT_xxx, if height is 0 */
	if(height == 0) {
		for(i = 0; i < scrinfo.fonts; ++i) {
			if(!strcmpi(pf[i].name, name)) {
				gr_fontsize = pf[i].cfont->height;
				return (PMWFONT)&pf[i];
			}
		}
first:
		/* return first builtin font*/
		gr_fontsize = pf->cfont->height;
		return (PMWFONT)&pf[0];
	}

#if HAVE_FREETYPE_SUPPORT
	pfont = (PMWFONT)freetype_createfont(name);
	if(pfont) {
		gr_fontsize = height;
		return pfont;
	} else printf("freetype_createfont: %s not found\n", name);
#endif

#if HAVE_T1LIB_SUPPORT
        /* Make sure the library is initialized */
        if(t1lib_init(psd)) {
		pfont = (PMWFONT)t1lib_createfont(name);
		if(pfont) {
			gr_fontsize = height;
			return pfont;
		} else printf("t1lib_createfont: %s not found\n", name);
	}
#endif

	/* find font closest in height*/
	if(height != 0) {
		fontno = 0;
		height = abs(height);
		fontht = COORD_MAX;
		for(i = 0; i < scrinfo.fonts; ++i) {
			pfont = (PMWFONT)&pf[i];
			GdGetFontInfo(pfont, &fontinfo);
			if(fontht > abs(height-fontinfo.height)) { 
				fontno = i;
				fontht = abs(height-fontinfo.height);
			}
		}
		gr_fontsize = pf[fontno].cfont->height;
		return (PMWFONT)&pf[fontno];
	}

	goto first;
}

/* unload and deallocate font*/
void
GdUnloadFont(PMWFONT pfont)
{
	pfont->procs->UnloadFont(pfont);
}

/* Return information about a specified font*/
BOOL
GdGetFontInfo(PMWFONT pfont, PFONTINFO pfontinfo)
{
	if(!pfont)
		return FALSE;

	return pfont->procs->GetFontInfo(pfont, pfontinfo);
}

/*
 * Convert from one encoding to another
 * Input cc and returned cc is character count, not bytes
 * NOTE: UTF8 not handled yet, need non-LGPL implementation
 */
int
GdConvertEncoding(const void *istr, int iflags, int cc, void *ostr, int oflags)
{
	const unsigned char 	*istr8 = istr;
	const unsigned short 	*istr16 = istr;
	const unsigned long	*istr32 = istr;
	unsigned char 		*ostr8 = ostr;
	unsigned short 		*ostr16 = ostr;
	unsigned long		*ostr32 = ostr;
	unsigned int		ch;
	int			icc;

	iflags &= TF_PACKMASK;
	oflags &= TF_PACKMASK;

	/* allow -1 for len with ascii*/
	if(cc == -1 && (iflags == TF_ASCII))
		cc = strlen(istr8);

	icc = cc;
	while(--icc >= 0) {
		switch(iflags) {
		default:
			ch = *istr8++;
			break;
		case TF_UC16:
			ch = *istr16++;
			break;
		case TF_UC32:
			ch = *istr32++;
		}
		switch(oflags) {
		default:
			*ostr8++ = (unsigned char)ch;
			break;
		case TF_UC16:
			*ostr16++ = (unsigned short)ch;
			break;
		case TF_UC32:
			*ostr32++ = ch;
		}
	}
	return cc;
}

/* Get the width and height of passed text string in the passed font*/
void
GdGetTextSize(PMWFONT pfont, const void *str, int cc, COORD *pwidth,
	COORD *pheight, COORD *pbase, int flags)
{
	const void *	text;
	unsigned long	buf[256];
	int		defencoding = gr_pfont->procs->encoding;

	/* convert encoding if required*/
	if((flags & TF_PACKMASK) != defencoding) {
		cc = GdConvertEncoding(str, flags, cc, buf, defencoding);
		flags &= ~TF_PACKMASK;
		flags |= defencoding;
		text = buf;
	} else text = str;

	/* calc height and width of string*/
	pfont->procs->GetTextSize(pfont, text, cc, pwidth, pheight, pbase);
}

/* Draw a text string at a specifed coordinates in the foreground color
 * (and possibly the background color), applying clipping if necessary.
 * The background color is only drawn if the gr_usebg flag is set.
 * Use the current font.
 */
void
GdText(PSD psd, COORD x, COORD y, const void *str, int cc, int flags)
{
	const void *	text;
	unsigned long	buf[256];
	int		defencoding = gr_pfont->procs->encoding;

	/* convert encoding if required*/
	if((flags & TF_PACKMASK) != defencoding) {
		cc = GdConvertEncoding(str, flags, cc, buf, defencoding);
		flags &= ~TF_PACKMASK;
		flags |= defencoding;
		text = buf;
	} else text = str;

	if(cc == 0)
		return;

	/* draw text string*/
	gr_pfont->procs->DrawText(gr_pfont, psd, x, y, text, cc, flags);
}

/*
 * Draw ascii text using COREFONT type font.
 */
void
corefont_drawtext(PMWFONT pfont, PSD psd, COORD x, COORD y,
	const void *text, int cc, int flags)
{
	const unsigned char *str = text;
	COORD		width;			/* width of text area */
	COORD 		height;			/* height of text area */
	COORD		base;			/* baseline of text*/
						/* bitmap for characters */
	IMAGEBITS     bitmap[MAX_CHAR_HEIGHT*MAX_CHAR_WIDTH/IMAGE_BITSPERIMAGE];

	pfont->procs->GetTextSize(pfont, str, cc, &width, &height, &base);
	
	if(flags & (TF_BOTTOM|TF_BASELINE))
		y -= (height - 1);

	switch (GdClipArea(psd, x, y, x + width - 1, y + height - 1)) {
	case CLIP_VISIBLE:
		/*
		 * For size considerations, there's no low-level text
		 * draw, so we've got to draw all text
		 * with per-point clipping for the time being
		if (gr_usebg)
			psd->FillRect(psd, x, y, x + width - 1, y + height - 1,
				gr_background);
		psd->DrawText(psd, x, y, str, cc, gr_foreground, pfont);
		GdFixCursor(psd);
		return;
		*/
		break;

	case CLIP_INVISIBLE:
		return;
	}

	/* Get the bitmap for each character individually, and then display
	 * them using clipping for each one.
	 */
	while (--cc >= 0 && x < psd->xvirtres) {
		pfont->procs->GetTextBits(pfont, *str++, bitmap, &width,
			&height, &base);
		/* note: change to bitmap*/
		GdBitmap(psd, x, y, width, height, bitmap);
		x += width;
	}
	GdFixCursor(psd);
}

#if HAVE_T1LIB_SUPPORT | HAVE_FREETYPE_SUPPORT
/*
 * Produce blend table from src and dst based on passed alpha table
 * Used because we don't quite yet have GdArea with alphablending,
 * so we pre-blend fg/bg colors for fade effect.
 */
static void
alphablend(PSD psd, OUTPIXELVAL *out, PIXELVAL src, PIXELVAL dst,
	unsigned char *alpha, int count)
{
	unsigned int	a, d;
	unsigned char	r, g, b;
	COLORVAL	palsrc, paldst;
	extern RGBENTRY gr_palette[256];

	while (--count >= 0) {
	    a = *alpha++;

#define BITS(pixel,shift,mask)	(((pixel)>>shift)&(mask))
	    if(a == 0)
		*out++ = dst;
	    else if(a == 255)
		*out++ = src;
	    else 
		switch(psd->pixtype) {
	        case PF_TRUECOLOR0888:
	        case PF_TRUECOLOR888:
		    d = BITS(dst, 16, 0xff);
		    r = (unsigned char)(((BITS(src, 16, 0xff) - d)*a)>>8) + d;
		    d = BITS(dst, 8, 0xff);
		    g = (unsigned char)(((BITS(src, 8, 0xff) - d)*a)>>8) + d;
		    d = BITS(dst, 0, 0xff);
		    b = (unsigned char)(((BITS(src, 0, 0xff) - d)*a)>>8) + d;
		    *out++ = (r << 16) | (g << 8) | b;
		    break;

	        case PF_TRUECOLOR565:
		    d = BITS(dst, 11, 0x1f);
		    r = (unsigned char)(((BITS(src, 11, 0x1f) - d)*a)>>8) + d;
		    d = BITS(dst, 5, 0x3f);
		    g = (unsigned char)(((BITS(src, 5, 0x3f) - d)*a)>>8) + d;
		    d = BITS(dst, 0, 0x1f);
		    b = (unsigned char)(((BITS(src, 0, 0x1f) - d)*a)>>8) + d;
		    *out++ = (r << 11) | (g << 5) | b;
		    break;

	        case PF_TRUECOLOR332:
		    d = BITS(dst, 5, 0x07);
		    r = (unsigned char)(((BITS(src, 5, 0x07) - d)*a)>>8) + d;
		    d = BITS(dst, 2, 0x07);
		    g = (unsigned char)(((BITS(src, 2, 0x07) - d)*a)>>8) + d;
		    d = BITS(dst, 0, 0x03);
		    b = (unsigned char)(((BITS(src, 0, 0x03) - d)*a)>>8) + d;
		    *out++ = (r << 5) | (g << 2) | b;
		    break;

	        case PF_PALETTE:
		    /* reverse lookup palette entry for blend ;-)*/
		    palsrc = GETPALENTRY(gr_palette, src);
		    paldst = GETPALENTRY(gr_palette, dst);
		    d = REDVALUE(paldst);
		    r = (unsigned char)(((REDVALUE(palsrc) - d)*a)>>8) + d;
		    d = GREENVALUE(paldst);
		    g = (unsigned char)(((GREENVALUE(palsrc) - d)*a)>>8) + d;
		    d = BLUEVALUE(paldst);
		    b = (unsigned char)(((BLUEVALUE(palsrc) - d)*a)>>8) + d;
		    *out++ = GdFindNearestColor(gr_palette, (int)psd->ncolors,
				RGB(r, g, b));
		    break;
	  	}
	}
}
#endif /*HAVE_T1LIB_SUPPORT | HAVE_FREETYPE_SUPPORT*/

#if HAVE_T1LIB_SUPPORT
/* contributed by Vidar Hokstad*/

static int
t1lib_init(PSD psd)
{
	static int inited = 0;

	if (inited)
		return 1;

	T1_SetBitmapPad(8);
	if (!T1_InitLib(0))
		return 0;
#ifdef T1LIB_USE_AA_HIGH	 
	T1_AASetLevel(T1_AA_HIGH);
#else
	T1_AASetLevel(T1_AA_LOW);
#endif	 
#if 0
	/* kluge: this is required if 16bpp drawarea driver is used*/
	if(psd->bpp == 16)
		T1_AASetBitsPerPixel(16);
	else
#endif
		T1_AASetBitsPerPixel(sizeof(PIXELVAL)*8);

	inited = 1;
	return 1;
}

static PMWT1LIBFONT
t1lib_createfont(const char *name)
{
	PMWT1LIBFONT	pf;
	int		id;
	char *		p;
	char		buf[256];

	/* match name against t1 font id's from t1 font database*/
	for(id=0; id<T1_Get_no_fonts(); ++id) {
		strcpy(buf, T1_GetFontFileName(id));

		/* remove extension*/
		for(p=buf; *p; ++p) {
			if(*p == '.') {
				*p = 0;
				break;
			}
		}

		if(!strcmpi(name, buf)) {
			/* allocate font structure*/
			pf = (PMWT1LIBFONT)calloc(sizeof(MWT1LIBFONT), 1);
			if (!pf)
				return NULL;
			pf->procs = &t1lib_procs;
			pf->fontid = id;
			return pf;
		}
	}
	return NULL;
}

/*
 * Draw ascii text string using T1LIB type font
 */
static void
t1lib_drawtext(PMWFONT pfont, PSD psd, COORD x, COORD y,
	const void *text, int cc, int flags)
{
	PMWT1LIBFONT	pf = (PMWT1LIBFONT)pfont;
	const unsigned char *str = text;
   	COORD		width;			/* width of text area */
	COORD 		height;			/* height of text area */
        GLYPH * g; /* T1lib glyph structure. Memory handling by T1lib */
#ifdef T1LIB_USE_AA_HIGH   
        OUTPIXELVAL	gvals[17];

        /* Blending array for antialiasing. The steeper the values increase
	 * near the end, the sharper the characters look, but also more jagged
	 */
        static unsigned char blend[17] = {
	   0x00, 0x00, 0x04, 0x0c, 0x10, 0x14, 0x18, 0x20,
	   0x30, 0x38, 0x40, 0x50, 0x70, 0x80, 0xa0, 0xc0, 0xff
	};
#else   
        OUTPIXELVAL	gvals[5];
        static unsigned char blend[5] = { 0x00, 0x44, 0x88, 0xcc, 0xff };
#endif   

        /* Check if we should throw out some fonts */

        if (gr_antialias) {
#ifdef T1LIB_USE_AA_HIGH      
	   alphablend(psd, gvals, gr_foreground, gr_background, blend, 17);
           T1_AAHSetGrayValues(gvals);
#else      
	   alphablend(psd, gvals, gr_foreground, gr_background, blend, 5);
           T1_AASetGrayValues(gvals[0],gvals[1],gvals[2],gvals[3],gvals[4]);
#endif
	   g = T1_AASetString(pf->fontid,(char *)str,cc,0, gr_kerning? T1_KERNING: 0,
			   gr_fontsize * 1.0, 0);

	   if (g && g->bits) {
	      //PIXELVAL save = gr_background;
	      width = g->metrics.rightSideBearing - g->metrics.leftSideBearing;
	      height = g->metrics.ascent - g->metrics.descent;

	      if(flags & (TF_BOTTOM|TF_BASELINE))
		y -= (height - 1);
	      
	      /* FIXME: Looks damn ugly if usebg is false. Will be handled when using
	       * alphablending in GdArea...
	       */
	      /* clipping handled in GdArea*/
	      /*FIXME kluge for transparency*/
	      //gr_background = gr_foreground + 1;
	      //gr_usebg = 0;
	      GdArea(psd,x,y, width, height, g->bits, PF_PIXELVAL);
	      //gr_background = save;
	   }
	} else {
	   /* Do non-aa drawing */
	   g = T1_SetString(pf->fontid,(char *)str,cc,0, gr_kerning? T1_KERNING: 0,
			gr_fontsize * 1.0, 0);

	   if (g && g->bits) {
	      unsigned char * b;
	      int xoff;
	      int maxy;
	      int xmod;
	      
	      /* I'm sure this sorry excuse for a bitmap rendering routine can
	       * be optimized quite a bit ;)
	       */
	      width = g->metrics.rightSideBearing - g->metrics.leftSideBearing;
	      height = g->metrics.ascent - g->metrics.descent;
	      if(flags & (TF_BOTTOM|TF_BASELINE))
		y -= (height - 1);
	      
	      b = g->bits;
	      maxy= y + height;
	      
/*	      if ((x + width) > psd->xvirtres) {
		 xmod = (x + width - psd->xvirtres + 7) >> 3;
		 width = width + x + width - psd->xvirtres;
	      } else xmod = 0;
*/
	      xmod = 0;
	      while (y < maxy) {
		 unsigned char data;
		 xoff = 0;
		 while (xoff < width ) {
		    if (!(xoff % 8)) {
		       data = *b;
		       b++;
		    }
		    
		    if (GdClipPoint(psd, x+xoff,y)) {
		       if (gr_usebg) {
	 		  psd->DrawPixel(psd,x+xoff,y,
			      data & (1 << (xoff % 8)) ?
			            gr_foreground : gr_background);
		       } else if (data & (1 << (xoff % 8))) {
			  psd->DrawPixel(psd,x+xoff,y, gr_foreground);
		       }
		    }
		    xoff++;
		 }
		 b += xmod;
		 y++;
	      }
	   }
        }

   if (g && g->bits) {
	   /* Save some memory */
	   free(g->bits);
           g->bits = 0; /* Make sure T1lib doesnt try to free it again */
   }

   GdFixCursor(psd);
}

static BOOL
t1lib_getfontinfo(PMWFONT pfont, PFONTINFO pfontinfo)
{
	int	i;
	COORD	width, height, baseline;

	/* FIXME, guess all sizes*/
	GdGetTextSize(pfont, "A", 1, &width, &height, &baseline, TF_ASCII);
	pfontinfo->height = height;
	pfontinfo->maxwidth = width;
	pfontinfo->baseline = 0;
	pfontinfo->firstchar = 32;
	pfontinfo->lastchar = 255;
	pfontinfo->fixed = TRUE;
	for(i=0; i<256; ++i)
		pfontinfo->widths[i] = width;
	return TRUE;
}

/* Get the width and height of passed text string in the current font*/
static void
t1lib_gettextsize(PMWFONT pfont, const void *text, int cc,
	COORD *pwidth, COORD *pheight, COORD *pbase)
{
	PMWT1LIBFONT		pf = (PMWT1LIBFONT)pfont;
	const unsigned char *	str = text;
	GLYPH *			g;

	g = T1_SetString(pf->fontid, (char *)str, cc, 0,
			gr_kerning? T1_KERNING: 0, gr_fontsize * 1.0, 0);
	*pwidth = g->metrics.rightSideBearing - g->metrics.leftSideBearing;
	*pheight = g->metrics.ascent - g->metrics.descent;
	if(g && g->bits) {
		free(g->bits);
		g->bits = 0;
	}
#if 0
	BBox 			b;

	/* FIXME. Something is *VERY* wrong here */
	b = T1_GetStringBBox(pf->fontid, str, cc, 0, gr_kerning?T1_KERNING:0);

	fprintf(stderr,"b.urx = %d, b.llx = %d\n",b.urx, b.llx);
	fprintf(stderr,"b.ury = %d, b.lly = %d\n",b.ury, b.lly);
	*pwidth = (b.urx - b.llx);
	*pheight = (b.lly - b.ury);
#endif
}

static void
t1lib_unloadfont(PMWFONT pfont)
{
	PMWT1LIBFONT	pf = (PMWT1LIBFONT)pfont;

	T1_DeleteAllSizes(pf->fontid);
	free(pf);
}

#endif /* HAVE_T1LIB_SUPPORT*/

#if HAVE_FREETYPE_SUPPORT
static PIXELVAL gray_palette[5];

static PMWFREETYPEFONT
freetype_createfont(const char *name)
{
	PMWFREETYPEFONT 	pf;
	unsigned short 		i, n;
	unsigned short 		platform, encoding;
	TT_Face_Properties 	properties;

	/* allocate font structure*/
	pf = (PMWFREETYPEFONT)calloc(sizeof(MWFREETYPEFONT), 1);
	if (!pf)
		return NULL;
	pf->procs = &freetype_procs;

	/* Init freetype library */
	if (TT_Init_FreeType (&pf->engine) != TT_Err_Ok) {
		free(pf);
		return NULL;
	}

	/* Init kerning extension */
	if (TT_Init_Kerning_Extension (pf->engine) != TT_Err_Ok)
		goto out;

	/* Load face */
	if (TT_Open_Face (pf->engine, name, &pf->face) != TT_Err_Ok)
		goto out;

	/* Load first kerning table */
	pf->can_kern = TRUE;
	if (TT_Load_Kerning_Table (pf->face, 0) != TT_Err_Ok)
		pf->can_kern = FALSE;
	else {
		if (TT_Get_Kerning_Directory (pf->face, &pf->directory)
		    != TT_Err_Ok)
			pf->can_kern = FALSE;
		else {
			/* Support only version 0 kerning table ... */
			if ((pf->directory.version != 0) ||
				(pf->directory.nTables <= 0) ||
				(pf->directory.tables->loaded != 1) ||
				(pf->directory.tables->version != 0) ||
				(pf->directory.tables->t.kern0.nPairs <= 0))
					pf->can_kern = FALSE;
		}
	}

	/* get face properties and allocate preload arrays */
	TT_Get_Face_Properties (pf->face, &properties);

	/* Create a glyph container */
	if (TT_New_Glyph (pf->face, &pf->glyph) != TT_Err_Ok)
		goto out;

	/* create instance */
	if (TT_New_Instance (pf->face, &pf->instance) != TT_Err_Ok)
		goto out;

	/* Set the instance resolution */
	if (TT_Set_Instance_Resolutions (pf->instance, 96, 96) != TT_Err_Ok)
		goto out;

	/* Look for a Unicode charmap */
	n = properties.num_CharMaps;

	for (i = 0; i < n; i++) {
		TT_Get_CharMap_ID (pf->face, i, &platform, &encoding);
		if (((platform == 3) && (encoding == 1)) ||
			((platform == 0) && (encoding == 0)))
		{
			TT_Get_CharMap (pf->face, i, &pf->char_map);
			i = n + 1;
		}
	}
	if (i == n) {
		printf("freetype_createfont: no unicode map table\n");
		goto out;
	}
	return pf;

out:
	TT_Done_FreeType (pf->engine);
	free(pf);
	return NULL;
}

static int
compute_kernval(PMWFREETYPEFONT pf, short current_glyph_code)
{
	int 		i = 0;
	int 		kernval;
	int 		nPairs = pf->directory.tables->t.kern0.nPairs;
	TT_Kern_0_Pair *pair = pf->directory.tables->t.kern0.pairs;

	if (pf->last_glyph_code != -1) {
		while ((pair->left != pf->last_glyph_code)
			&& (pair->right != current_glyph_code))
		{
			pair++;
			i++;
			if (i == nPairs)
			break;
		}

		if (i == nPairs)
			kernval = 0;
		else
			/* FIXME is +32 right?*/
			kernval = (pair->value + 32) & -64;
	} else
		kernval = 0;

	return kernval;
}


/* Render a single glyph*/
static void
drawchar(PMWFREETYPEFONT pf, PSD psd, TT_Glyph glyph, int x_offset,
	int y_offset)
{
	TT_F26Dot6 	xmin, ymin, xmax, ymax, x, y, z;
	unsigned char 	*src, *srcptr;
	PIXELVAL 	*dst, *dstptr;
	PIXELVAL 	*bitmap;
	int 		size, width, height;
	TT_Outline 	outline;
	TT_BBox 	bbox;
	TT_Raster_Map 	Raster;
	TT_Error 	error;
	//PIXELVAL 	save;

	/* we begin by grid-fitting the bounding box */
	TT_Get_Glyph_Outline (pf->glyph, &outline);
	TT_Get_Outline_BBox (&outline, &bbox);

	xmin = (bbox.xMin & -64) >> 6;
	ymin = (bbox.yMin & -64) >> 6;
	xmax = ((bbox.xMax + 63) & -64) >> 6;
	ymax = ((bbox.yMax + 63) & -64) >> 6;
	width = xmax - xmin;
	height = ymax - ymin;
	size = width * height;

	/* now re-allocate the raster bitmap */
	Raster.rows = height;
	Raster.width = width;

	if (gr_antialias)
		Raster.cols = (Raster.width + 3) & -4;	/* pad to 32-bits */
	else
		Raster.cols = (Raster.width + 7) & -8;	/* pad to 64-bits ??? */

	Raster.flow = TT_Flow_Up;
	Raster.size = Raster.rows * Raster.cols;
	Raster.bitmap = malloc (Raster.size);

	memset (Raster.bitmap, 0, Raster.size);

	/* now render the glyph in the small pixmap */

	/* IMPORTANT NOTE: the offset parameters passed to the function     */
	/* TT_Get_Glyph_Bitmap() must be integer pixel values, i.e.,        */
	/* multiples of 64.  HINTING WILL BE RUINED IF THIS ISN'T THE CASE! */
	/* This is why we _did_ grid-fit the bounding box, especially xmin  */
	/* and ymin.                                                        */

	if (!gr_antialias)
		error = TT_Get_Glyph_Bitmap (pf->glyph, &Raster,
				-xmin * 64, -ymin * 64);
	else
		error = TT_Get_Glyph_Pixmap (pf->glyph, &Raster,
				-xmin * 64, -ymin * 64);

	if (error) {
		free (Raster.bitmap);
		return;
	}

	bitmap = malloc (size * sizeof (PIXELVAL));
	memset (bitmap, 0, size * sizeof (PIXELVAL));

	src = (char *) Raster.bitmap;
	dst = bitmap + (size - width);

	for (y = ymin; y < ymax; y++) {
		srcptr = src;
		dstptr = dst;

		for (x = xmin; x < xmax; x++) {
			if (gr_antialias)
				*dstptr++ = gray_palette[(int) *srcptr];
			else {
				for(z=0;
				    z <= ((xmax-x-1) < 7 ? (xmax-x-1) : 7);
				    z++) {
					*dstptr++ = ((*srcptr << z) & 0x80)?
						gr_foreground: gr_background;
				}
				x += 7;
			}

			srcptr++;
		}

		src += Raster.cols;
		dst -= width;
	}

	/* FIXME - must clear background upstairs if not gr_usebg*/
	/* FIXME: GdArea problem if fg == bg*/
	//save = gr_background;
	//gr_background = gr_foreground + 1;

	/* Now draw the bitmap ... */
	GdArea(psd, x_offset + xmin, y_offset - (ymin + height), width, height,
		bitmap, PF_PIXELVAL);

	//gr_background = save;

	free (bitmap);
	free (Raster.bitmap);
}

/*
 * Draw unicode 16 text string using FREETYPE type font
 */
static void
freetype_drawtext(PMWFONT pfont, PSD psd, COORD ax, COORD ay,
	const void *text, int cc, int flags)
{
	PMWFREETYPEFONT	pf = (PMWFREETYPEFONT)pfont;
	const unsigned short *	str = text;
	TT_F26Dot6 	x = ax, y = ay;
	TT_Pos 		vec_x, vec_y;
	int 		i;
	float 		angle;
	TT_Outline 	outline;
	TT_UShort 	curchar;
	TT_Glyph_Metrics metrics;
	static unsigned char blend[5] = { 0x00, 0x44, 0x88, 0xcc, 0xff };
	static unsigned char virtual_palette[5] = { 0, 1, 2, 3, 4 };

	if(!(flags & (TF_BOTTOM|TF_BASELINE)))
		y += gr_fontsize;

	pf->last_glyph_code = -1;		/* reset kerning*/
	pf->last_pen_pos = -1;
	TT_Set_Instance_CharSize (pf->instance, gr_fontsize * 64);

	/* Build the rotation matrix with the given angle */
	if (gr_rotation) {
		/* FIXME: what if rotation goes from nonzero to zero?*/
		TT_Set_Instance_Transform_Flags (pf->instance, TRUE, FALSE);

		angle = gr_rotation * M_PI / 180;
		pf->matrix.yy = (TT_Fixed) (cos (angle) * (1 << 16));
		pf->matrix.yx = (TT_Fixed) (sin (angle) * (1 << 16));
		pf->matrix.xx = pf->matrix.yy;
		pf->matrix.xy = -pf->matrix.yx;
	}

	/* Set the "graylevels" */
	if (gr_antialias)
		TT_Set_Raster_Gray_Palette (pf->engine, virtual_palette);

	if(gr_antialias)
		alphablend(psd, gray_palette, gr_foreground, gr_background,
			blend, 5);

	for (i = 0; i < cc; i++) {
		curchar = TT_Char_Index (pf->char_map, str[i]);

		if (TT_Load_Glyph (pf->instance, pf->glyph, curchar,
			TTLOAD_DEFAULT) != TT_Err_Ok)
				continue;

		if (gr_rotation) {
			TT_Get_Glyph_Outline (pf->glyph, &outline);
			TT_Transform_Outline (&outline, &pf->matrix);
		}

		TT_Get_Glyph_Metrics (pf->glyph, &metrics);

		if (gr_kerning && pf->can_kern) {
			if (gr_rotation) {
				vec_x = compute_kernval(pf, curchar);
				vec_y = 0;
				TT_Transform_Vector(&vec_x, &vec_y,&pf->matrix);

				x += vec_x / 64;
				y -= vec_y / 64;
			} else
				x += compute_kernval(pf, curchar) / 64;
		}

		drawchar(pf, psd, pf->glyph, x, y);

		if (gr_rotation) {
			vec_x = metrics.advance;
			vec_y = 0;
			TT_Transform_Vector (&vec_x, &vec_y, &pf->matrix);

			x += vec_x / 64;
			y -= vec_y / 64;
		} else {
			x += metrics.advance / 64;

			/* Kerning point syndrome avoidance */
			if (pf->last_pen_pos > x)
				x = pf->last_pen_pos;
			pf->last_pen_pos = x;
		}

		pf->last_glyph_code = curchar;
	}
}

/*
 * Return information about a specified font.
 */
static BOOL
freetype_getfontinfo(PMWFONT pfont, PFONTINFO pfontinfo)
{
	int	i;
	COORD	width, height, baseline;

	/* FIXME, guess all sizes*/
	GdGetTextSize(pfont, "A", 1, &width, &height, &baseline, TF_ASCII);
	pfontinfo->height = height;
	pfontinfo->maxwidth = width;
	pfontinfo->baseline = 0;
	pfontinfo->firstchar = 32;
	pfontinfo->lastchar = 255;
	pfontinfo->fixed = TRUE;
	for(i=0; i<256; ++i)
		pfontinfo->widths[i] = width;
	return TRUE;
}

static void
freetype_gettextsize(PMWFONT pfont, const void *text, int cc,
	COORD *pwidth, COORD *pheight, COORD *pbase)
{
#if 0 // FIXME rewrite
	PMWFREETYPEFONT	pf = (PMWFREETYPEFONT)pfont;
	const unsigned short *	str = text;
	unsigned short	c;
	int		width;

	if(pf->width == NULL)
		width = cc * pf->maxwidth;
	else {
		width = 0;
		while(--cc >= 0) {
			if( (c = *str++) >= pf->firstchar &&
			     c < pf->firstchar+pf->size)
				width += pf->width[c - pf->firstchar];
		}
	}
	*pwidth = width;
	*pheight = pf->height;
	*pbase = 0;
#endif
	*pwidth = 10;
	*pheight = 10;
	*pbase = 0;
}

static void
freetype_unloadfont(PMWFONT pfont)
{
	PMWFREETYPEFONT	pf = (PMWFREETYPEFONT)pfont;

	TT_Done_FreeType(pf->engine);
	free(pf);
}
#endif /* HAVE_FREETYPE_SUPPORT*/
