/*
 * Copyright (c) 1999 Greg Haerr <greg@censoft.com>
 * Copyright (c) 1999 Alex Holden <alex@linuxhacker.org>
 * Portions Copyright (c) 1991 David I. Bell
 *
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 *
 * Completely rewritten for speed by Greg Haerr
 *
 * This is the server side of the network interface, which accepts
 * connections from clients, receives functions from them, and dispatches
 * events to them.
 */
#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/socket.h>
#if ELKS
#include <linuxmt/na.h>
#else
#include <sys/un.h>
#endif
#include "serv.h"
#include "nxproto.h"

extern	int		un_sock;
extern	GR_CLIENT	*root_client;
extern	int		current_fd;

/*
 * Wrapper functions called after full packet read
 */
void 
GrOpenWrapper(void *r)
{
	GrOpen();
}

void
GrCloseWrapper(void *r)
{
	GrClose();
}

void
GrGetScreenInfoWrapper(void *r)
{
	GR_SCREEN_INFO si;

	GrGetScreenInfo(&si);
	GsWrite(current_fd, &si, sizeof(si));
}

void
GrNewWindowWrapper(void *r)
{
	nxNewWindowReq *req = r;
	GR_WINDOW_ID	wid;

	wid = GrNewWindow(req->parentid, req->x, req->y, req->width,
		req->height, req->bordersize, req->backgroundcolor,
		req->bordercolor);

	GsWrite(current_fd, &wid, sizeof(wid));
}


void
GrNewPixmapWrapper(void *r)
{
	nxNewPixmapReq *req = r;
	GR_WINDOW_ID	wid;

        /* FIXME: Add support for passing info about shared memory
	 * segment
	 */
	wid = GrNewPixmap(req->width, req->height, 0);

	GsWrite(current_fd, &wid, sizeof(wid));
}

void
GrNewInputWindowWrapper(void *r)
{
	nxNewInputWindowReq *req = r;
	GR_WINDOW_ID         wid;

	wid = GrNewInputWindow(req->parentid, req->x, req->y, req->width,
		req->height);
	GsWrite(current_fd, &wid, sizeof(wid));
}

void
GrDestroyWindowWrapper(void *r)
{
	nxDestroyWindowReq *req = r;

	GrDestroyWindow(req->windowid);
}

void
GrNewGCWrapper(void *r)
{
	GR_GC_ID gc = GrNewGC();

	GsWrite(current_fd, &gc, sizeof(gc));
}

void
GrCopyGCWrapper(void *r)
{
	nxCopyGCReq *req = r;
	GR_GC_ID     gcnew;

	gcnew = GrCopyGC(req->gcid);
	GsWrite(current_fd, &gcnew, sizeof(gcnew));
}

void
GrGetGCInfoWrapper(void *r)
{
	nxGetGCInfoReq *req = r;
	GR_GC_INFO      gcinfo;

	GrGetGCInfo(req->gcid, &gcinfo);
	GsWrite(current_fd, &gcinfo, sizeof(gcinfo));
}

void
GrDestroyGCWrapper(void *r)
{
	nxDestroyGCReq *req = r;

	GrDestroyGC(req->gcid);
}

void
GrMapWindowWrapper(void *r)
{
	nxMapWindowReq *req = r;

	GrMapWindow(req->windowid);
}

void
GrUnmapWindowWrapper(void *r)
{
	nxUnmapWindowReq *req = r;

	GrUnmapWindow(req->windowid);
}

void
GrRaiseWindowWrapper(void *r)
{
	nxRaiseWindowReq *req = r;

	GrRaiseWindow(req->windowid);
}

void
GrLowerWindowWrapper(void *r)
{
	nxLowerWindowReq *req = r;

	GrLowerWindow(req->windowid);
}

void
GrMoveWindowWrapper(void *r)
{
	nxMoveWindowReq *req = r;

	GrMoveWindow(req->windowid, req->x, req->y);
}

void
GrResizeWindowWrapper(void *r)
{
	nxResizeWindowReq *req = r;

	GrResizeWindow(req->windowid, req->width, req->height);
}

void
GrReparentWindowWrapper(void *r)
{
	nxReparentWindowReq *req = r;

	GrReparentWindow(req->windowid, req->parentid, req->x, req->y);
}

void
GrGetWindowInfoWrapper(void *r)
{
	nxGetWindowInfoReq *req = r;
	GR_WINDOW_INFO      wi;

	GrGetWindowInfo(req->windowid, &wi);
	GsWrite(current_fd, &wi, sizeof(wi));
}

void
GrGetFontInfoWrapper(void *r)
{
	nxGetFontInfoReq *req = r;
	GR_FONT_INFO      fi;

	GrGetFontInfo(req->fontid, &fi);
	GsWrite(current_fd, &fi, sizeof(fi));
}

void
GrSetFocusWrapper(void *r)
{
	nxSetFocusReq *req = r;

	GrSetFocus(req->windowid);
}

void
GrSetBorderColorWrapper(void *r)
{
	nxSetBorderColorReq *req = r;

	GrSetBorderColor(req->windowid, req->color);
}

void
GrClearWindowWrapper(void *r)
{
	nxClearWindowReq *req = r;

	GrClearWindow(req->windowid, req->exposeflag);
}

void
GrSelectEventsWrapper(void *r)
{
	nxSelectEventsReq *req = r;

	GrSelectEvents(req->windowid, req->eventmask);
}

void
GrGetNextEventWrapper(void *r)
{
#if 1
	/* tell main loop to call Finish routine on event*/
	curclient->waiting_for_event = TRUE;
#else
	GR_EVENT evt;

	/* first check if any event ready*/
	GrCheckNextEvent(&evt);
	if(evt.type == GR_EVENT_TYPE_NONE) {
		/* tell main loop to call Finish routine on event*/
		curclient->waiting_for_event = TRUE;
		return;
	}

	GsWrite(current_fd, &evt, sizeof(evt));
#endif
}

/* Complete the GrGetNextEvent call from client.
 * The client is still waiting on a read at this point.
 */
void
GrGetNextEventWrapperFinish(int fd)
{
	GR_EVENT evt;

	/* get the event and pass it to client*/
	/* this will never be GR_EVENT_TYPE_NONE*/
	GrCheckNextEvent(&evt);

	GsWrite(fd, &evt, sizeof(evt));
}

void
GrCheckNextEventWrapper(void *r)
{
	GR_EVENT evt;

	GrCheckNextEvent(&evt);

	GsWrite(current_fd, &evt, sizeof(evt));
}

void
GrPeekEventWrapper(void *r)
{
	GR_EVENT evt;
	GR_CHAR	 ret;

	ret = GrPeekEvent(&evt);

	GsWrite(current_fd, &evt, sizeof(evt));
	GsWrite(current_fd, &ret, sizeof(GR_CHAR));
}

void
GrLineWrapper(void *r)
{
	nxLineReq *req = r;

	GrLine(req->drawid, req->gcid, req->x1, req->y1, req->x2, req->y2);
}

void
GrPointWrapper(void *r)
{
	nxPointReq *req = r;

	GrPoint(req->drawid, req->gcid, req->x, req->y);

}

void
GrRectWrapper(void *r)
{
	nxRectReq *req = r;

	GrRect(req->drawid, req->gcid, req->x, req->y, req->width, req->height);
}

void
GrFillRectWrapper(void *r)
{
	nxFillRectReq *req = r;

	GrFillRect(req->drawid, req->gcid, req->x, req->y, req->width,
		req->height);
}

void
GrPolyWrapper(void *r)
{
	nxPolyReq *req = r;
	int        count;

	count = GetReqVarLen(req) / sizeof(GR_POINT);
	GrPoly(req->drawid, req->gcid, count, (GR_POINT *)GetReqData(req));
}

/* FIXME: fails with pointtable size > 64k if sizeof(int) == 2*/
void
GrFillPolyWrapper(void *r)
{
	nxPolyReq *req = r;
	int        count;

	count = GetReqVarLen(req) / sizeof(GR_POINT);
	GrFillPoly(req->drawid, req->gcid, count, (GR_POINT *)GetReqData(req));
}

void
GrEllipseWrapper(void *r)
{
	nxEllipseReq *req = r;

	GrEllipse(req->drawid, req->gcid, req->x, req->y, req->rx, req->ry);
}

void
GrFillEllipseWrapper(void *r)
{
	nxFillEllipseReq *req = r;

	GrFillEllipse(req->drawid, req->gcid, req->x, req->y, req->rx, req->ry);
}

void
GrSetGCForegroundWrapper(void *r)
{
	nxSetGCForegroundReq *req = r;

	GrSetGCForeground(req->gcid, req->color);
}

void
GrSetGCBackgroundWrapper(void *r)
{
	nxSetGCBackgroundReq *req = r;

	GrSetGCBackground(req->gcid, req->color);
}

void
GrSetGCUseBackgroundWrapper(void *r)
{
	nxSetGCUseBackgroundReq *req = r;

	GrSetGCUseBackground(req->gcid, req->flag);
}

void
GrSetGCModeWrapper(void *r)
{
	nxSetGCModeReq *req = r;

	GrSetGCMode(req->gcid, req->mode);
}

void
GrCreateFontWrapper(void *r)
{
	nxCreateFontReq *req = r;
	GR_FONTID 	fontid;
	char *		p;

	p = GetReqData(req);
	p[GetReqVarLen(req)] = 0;
	fontid = GrCreateFont(p, req->height);
	GsWrite(current_fd, &fontid, sizeof(fontid));
}

void
GrUnloadFontWrapper(void *r)
{
	nxUnloadFontReq *req = r;

	GrUnloadFont(req->fontid);
}

void
GrSetGCFontWrapper(void *r)
{
	nxSetGCFontReq *req = r;

	GrSetGCFont(req->gcid, req->fontid);
}

void
GrSetGCFontSizeWrapper(void *r)
{
	nxSetGCFontSizeReq *req = r;

 	GrSetGCFontSize(req->gcid, req->fontsize);
}

void
GrGetGCTextSizeWrapper(void *r)
{
	nxGetGCTextSizeReq *req = r;
	GR_SIZE             retwidth, retheight, retbase;

	GrGetGCTextSize(req->gcid, GetReqData(req), GetReqVarLen(req),
		req->flags, &retwidth, &retheight, &retbase);

	GsWrite(current_fd, &retwidth, sizeof(retwidth));
	GsWrite(current_fd, &retheight, sizeof(retheight));
	GsWrite(current_fd, &retbase, sizeof(retbase));
}

/* FIXME: fails with size > 64k if sizeof(int) == 2*/
void
GsReadAreaWrapper(void *r)
{
	nxReadAreaReq *req = r;
	int            size;
	PIXELVAL *     area;

	/* FIXME: optimize for smaller pixelvals*/
	size = req->width * req->height * sizeof(PIXELVAL);

	if(!(area = malloc(size))) {
		/*GsPutCh(current_fd, GrRetENoMem);*/ /* FIXME*/
		/*return;*/
	}

	GrReadArea(req->drawid, req->x, req->y, req->width, req->height, area);
	GsWrite(current_fd, area, size);
	free(area);
}

/* FIXME: fails with size > 64k if sizeof(int) == 2*/
void
GrAreaWrapper(void *r)
{
	nxAreaReq *req = r;

	GrArea(req->drawid, req->gcid, req->x, req->y, req->width, req->height,
		GetReqData(req), req->pixtype);
}

/* FIXME: fails with bitmapsize > 64k if sizeof(int) == 2*/
void
GrBitmapWrapper(void *r)
{
	nxBitmapReq *req = r;

	GrBitmap(req->drawid, req->gcid, req->x, req->y, req->width,
		req->height, GetReqData(req));
}


void
GrCopyAreaWrapper(void *r)
{
	nxCopyAreaReq *req = r;

	GrCopyArea(req->drawid, req->gcid, req->x, req->y, req->width, req->height,
		req->srcid, req->srcx, req->srcy, req->op);
}

void
GrTextWrapper(void *r)
{
	nxTextReq *req = r;

	GrText(req->drawid, req->gcid, req->x, req->y, GetReqData(req),
		req->count, req->flags);
}

void
GrSetCursorWrapper(void *r)
{
	nxSetCursorReq *req = r;
	int		varlen;
	char *          bitmap;

	varlen = GetReqVarLen(req);
	bitmap = GetReqData(req);

	GrSetCursor(req->windowid, req->width, req->height,
		req->hotx, req->hoty, req->fgcolor, req->bgcolor, 
		(GR_BITMAP *)bitmap, (GR_BITMAP *)(bitmap+varlen/2));
}

void
GrMoveCursorWrapper(void *r)
{
	nxMoveCursorReq *req = r;

	GrMoveCursor(req->x, req->y);
}

void
GrGetSystemPaletteWrapper(void *r)
{
	GR_PALETTE             pal;

	GrGetSystemPalette(&pal);
	GsWrite(current_fd, &pal, sizeof(pal));
}

void
GrFindColorWrapper(void *r)
{
	nxFindColorReq *req = r;
	PIXELVAL	pixel;

	GrFindColor(req->color, &pixel);
	GsWrite(current_fd, &pixel, sizeof(pixel));
}

void
GrJPEGWrapper(void *r)
{
	nxJPEGReq *req = r;

	GrJPEG(req->drawid, req->gcid, req->x, req->y, req->width,
		req->height, req->fast_grayscale, GetReqData(req));
}

void
GrBMPWrapper(void *r)
{
	nxBMPReq *req = r;

	GrBMP(req->drawid, req->gcid, req->x, req->y, req->width,
		req->height, GetReqData(req));
}

/*
 * Handler functions, ordered by reqType
 */
struct GrFunction {
	void		(*func)(void *);
	GR_FUNC_NAME 	name;
} GrFunctions[] = {
	{GrOpenWrapper, "GrOpen"},
	{GrCloseWrapper, "GrClose"},
	{GrGetScreenInfoWrapper, "GrGetScreenInfo"},
	{GrNewWindowWrapper, "GrNewWindow"},
	{GrNewInputWindowWrapper, "GrNewInputWindow"},
	{GrDestroyWindowWrapper, "GrDestroyWindow"},
	{GrNewGCWrapper, "GrNewGC"},
	{GrCopyGCWrapper, "GrCopyGC"},
	{GrGetGCInfoWrapper, "GrGetGCInfo"},
	{GrDestroyGCWrapper, "GrDestroyGC"},
	{GrMapWindowWrapper, "GrMapWindow"},
	{GrUnmapWindowWrapper, "GrUnmapWindow"},
	{GrRaiseWindowWrapper, "GrRaiseWindow"},
	{GrLowerWindowWrapper, "GrLowerWindow"},
	{GrMoveWindowWrapper, "GrMoveWindow"},
	{GrResizeWindowWrapper, "GrResizeWindow"},
	{GrGetWindowInfoWrapper, "GrGetWindowInfo"},
	{GrGetFontInfoWrapper, "GrGetFontInfo"},
	{GrSetFocusWrapper, "GrSetFocus"},
	{GrSetBorderColorWrapper, "GrSetBorderColor"},
	{GrClearWindowWrapper, "GrClearWindowWrapper"},
	{GrSelectEventsWrapper, "GrSelectEvents"},
	{GrGetNextEventWrapper, "GrGetNextEvent"},
	{GrCheckNextEventWrapper, "GrCheckNextEvent"},
	{GrPeekEventWrapper, "GrPeekEvent"},
	{GrLineWrapper, "GrLine"},
	{GrPointWrapper, "GrPoint"},
	{GrRectWrapper, "GrRect"},
	{GrFillRectWrapper, "GrFillRect"},
	{GrPolyWrapper, "GrPoly"},
	{GrFillPolyWrapper, "GrFillPoly"},
	{GrEllipseWrapper, "GrEllipse"},
	{GrFillEllipseWrapper, "GrFillEllipse"},
	{GrSetGCForegroundWrapper, "GrSetGCForeground"},
	{GrSetGCBackgroundWrapper, "GrSetGCBackGround"},
	{GrSetGCUseBackgroundWrapper, "GrSetGCUseBackGround"},
	{GrSetGCModeWrapper, "GrSetGCMode"},
	{GrSetGCFontWrapper, "GrSetGCFont"},
	{GrGetGCTextSizeWrapper, "GrGetGCTextSize"},
	{GsReadAreaWrapper, "GsReadArea"},
	{GrAreaWrapper, "GrArea"},
	{GrBitmapWrapper, "GrBitmap"},
	{GrTextWrapper, "GrText"},
	{GrSetCursorWrapper, "GrSetCursor"},
	{GrMoveCursorWrapper, "GrMoveCursor"},
	{GrGetSystemPaletteWrapper, "GrGetSystemPalette"},
	{GrFindColorWrapper, "GrFindColor"},
	{GrReparentWindowWrapper, "GrReparentWindow"},
 	{GrJPEGWrapper, "GrJPEG"},
 	{GrBMPWrapper, "GrBMP"},
        {GrNewPixmapWrapper, "GrNewPixmap"},
        {GrCopyAreaWrapper, "GrCopyArea"},
        {GrSetGCFontSizeWrapper, "GrSetGCFontSize"},
        {GrCreateFontWrapper, "GrCreateFont"},
        {GrUnloadFontWrapper, "GrUnloadFont"},
};

/*
 * This function is used to bind to the named socket which is used to
 * accept connections from the clients.
 */
int 
GsOpenSocket(void)
{
	struct stat s;
#if ELKS
	struct sockaddr_na sckt;
#ifndef SUN_LEN
#define SUN_LEN(ptr)	(sizeof(sckt))
#endif
#else
	struct sockaddr_un sckt;
#ifndef SUN_LEN
#define SUN_LEN(ptr)	((size_t) (((struct sockaddr_un *) 0)->sun_path) \
		      		+ strlen ((ptr)->sun_path))
#endif
#endif /* ELKS */

#if ELKS
	if((un_sock = socket(AF_NANO, SOCK_STREAM, 0)) == -1)
		return -1;

	sckt.sun_family = AF_NANO;
	sckt.sun_no = GR_NUMB_SOCKET;
#else
	/* Check if the file already exists: */
	if(!stat(GR_NAMED_SOCKET, &s)) {
		/* FIXME: should try connecting to see if server is active */
		if(unlink(GR_NAMED_SOCKET))
			return -1;
	}

	/* Create the socket: */
	if((un_sock = socket(AF_UNIX, SOCK_STREAM, 0)) == -1)
		return -1;

	/* Bind a name to the socket: */
	sckt.sun_family = AF_UNIX;
	strncpy(sckt.sun_path, GR_NAMED_SOCKET, sizeof(sckt.sun_path));
#endif /* ELKS */
	if(bind(un_sock, (struct sockaddr *) &sckt, SUN_LEN(&sckt)) < 0)
		return -1;

	/* Start listening on the socket: */
	if(listen(un_sock, 5) == -1)
		return -1;
	return 1;
}

void
GsCloseSocket(void)
{
	if(un_sock != -1)
		close(un_sock);
	un_sock = -1;
	unlink(GR_NAMED_SOCKET);
}

/*
 * This function is used to accept a connnection from a client.
 */
void
GsAcceptClient(void)
{
	int i;
#if ELKS
	struct sockaddr_na sckt;
#else
	struct sockaddr_un sckt;
#endif
	size_t size = sizeof(sckt);

	if((i = accept(un_sock, (struct sockaddr *) &sckt, &size)) == -1) {
		printf("accept failed (%d)\n", errno);
		return;
	}
	GsAcceptClientFd(i);
}

/*
 * This function accepts a client ID, and searches through the
 * linked list of clients, returning a pointer to the relevant
 * structure or NULL if not found.
 */
GR_CLIENT *
GsFindClient(int fd)
{
	GR_CLIENT *client;

	client = root_client;

	while(client) {
		if(client->id == fd) return(client);
		client = client->next;
	}
	
	return 0;
}

/*
 * Destroy windows and eventclient structures used by client.
 * Called by GsDropClient after a client has exited to clean
 * up resources.
 */
void
GsDestroyClientResources(GR_CLIENT * client)
{
	GR_WINDOW     * wp;
	GR_WINDOW     * nwp;
	GR_PIXMAP     * pp;
	GR_PIXMAP     * npp;
	GR_EVENT_CLIENT *ecp;
	GR_EVENT_CLIENT *necp;
	GR_EVENT_CLIENT *pecp = NULL;

	/* search window list, destroy windows owned by client*/
	for(wp=listwp; wp; wp=nwp) {
		nwp = wp->next;
		if (wp->owner == client) {
			/* FIXME: change to call GsDestroyWindow()*/
 			GsWpUnmapWindow(wp);
			GsDeliverUpdateEvent(wp, GR_UPDATE_UNMAP, 0, 0, 0, 0);
			GsWpDestroyWindow(wp);
		} else {
			/*
			 * Remove eventclient structures for
			 * windows this client didn't own
			 */
			ecp = wp->eventclients;
			while (ecp) {
				necp = ecp->next;
				if (ecp->client == client) {
					if (ecp == wp->eventclients)
						wp->eventclients = ecp->next;
					else
						pecp->next = ecp->next;
					free(ecp);
				} else
					pecp = ecp;
				ecp = necp;
			}
		}
	}

	/* search pixmap list, destroy pixmaps owned by client*/
	for(pp=listpp; pp; pp=npp) {
		npp = pp->next;
		if (pp->owner == client)
			GrDestroyWindow(pp->id);
	}
}

/*
 * This is used to drop a client when it is detected that the connection to it
 * has been lost.
 */
void
GsDropClient(int fd)
{
	GR_CLIENT *client;

	close(fd);	/* Close the socket */

	if((client = GsFindClient(fd))) { /* If it exists */
		GsDestroyClientResources(client);
		if(client == root_client)
			root_client = client->next;
		/* Link the prev to the next */
		if(client->prev) client->prev->next = client->next;

		/* Link the next to the prev */
		if(client->next) client->next->prev = client->prev;

		free(client);	/* Free the structure */
	} else fprintf(stderr, "Error: trying to drop non-existent client %d.\n", fd);
}

/*
 * This is a wrapper to read() which handles error conditions, and
 * returns 0 for both error conditions and no data.
 */
int
#if ELKS
GsRead(int fd, char *buf, int c)
#else
GsRead(int fd, void *buf, int c)
#endif
{
	int e, n;

	n = 0;

	while(n < c) {
		e = read(fd, (buf + n), (c - n));
		if(e <= 0) {
printf("GsRead: read failed %d %x %d: %d\r\n", e, buf+n, c-n, errno);
			GrClose();
			return -1;
		}
		n += e;
	}

	return 0;
}

#if 0
int GsPutCh(int fd, unsigned char c)
{
	return(GsWrite(fd, &c, 1));
}
#endif

/*
 * This is a wrapper to write().
 */
int GsWrite(int fd, void *buf, int c)
{
	int e, n;

	n = 0;

	while(n < c) {
		e = write(fd, ((char *) buf + n), (c - n));
		if(e <= 0) {
			GrClose();
			return -1;
		}
		n += e;
	}

	return 0;
}

/*
 * This function is used to parse and dispatch requests from the clients.
 * Note that the maximum request size is allocated from the stack
 * in this function.
 */
void
GsHandleClient(int fd)
{
	nxReq *	req;
	long	len;
	char	buf[MAXREQUESTSZ];

	current_fd = fd;

	/* read request header*/
	if(GsRead(fd, buf, sizeof(nxReq)))
		return;

	len = GetReqAlignedLen((nxReq *)&buf[0]);
	if(len > sizeof(nxReq)) {
		if(len > MAXREQUESTSZ) {
			printf("GsHandleClient: request too large: %ld > %d\n",
				len, MAXREQUESTSZ);
			exit(1);
		}
		/* read additional request data*/
		if(GsRead(fd, &buf[sizeof(nxReq)], len-sizeof(nxReq)))
			return;
	}
	req = (nxReq *)&buf[0];

	if(req->reqType < GrTotalNumCalls) {
		curfunc = GrFunctions[req->reqType].name;
/*printf("HandleClient %s\r\n", curfunc);*/
		GrFunctions[req->reqType].func(req);
	} else {
		printf("Bad function\n");
	}
}
